"""copyright (c) 2021 Beeflow Ltd.

Author Rafal Przetakowski <rafal.p@beeflow.co.uk>"""
import json
import sys
from io import StringIO
from unittest import TestCase

from unittest_data_provider import data_provider

from beeflow_ajax.lib import AjaxResponse


class TestAjaxResponse(TestCase):
    datasource_for_tests = lambda: (  # noqa: E731
        # ("function name", {parameters to unpack}, expected result),
        ("alert", {"msg": "team"}, [{"cmd": AjaxResponse.ALERT, "data": "team"}]),
        (
            "alert_success",
            {"msg": "team", "title": "other"},
            [{"cmd": AjaxResponse.ALERT_SUCCESS, "data": "team", "title": "other"}],
        ),
        (
            "alert_error",
            {"msg": "team", "title": "other"},
            [{"cmd": AjaxResponse.ALERT_ERROR, "data": "team", "title": "other"}],
        ),
        (
            "alert_warning",
            {"msg": "team", "title": "other"},
            [{"cmd": AjaxResponse.ALERT_WARNING, "data": "team", "title": "other"}],
        ),
        (
            "alert_info",
            {"msg": "team", "title": "other"},
            [{"cmd": AjaxResponse.ALERT_INFO, "data": "team", "title": "other"}],
        ),
        (
            "debug",
            {"data": {"team": "team name", "title": "other"}},
            [
                {
                    "cmd": AjaxResponse.DEBUG,
                    "data": {"team": "team name", "title": "other"},
                }
            ],
        ),
        (
            "append",
            {"element": "#team", "value": "some value"},
            [{"cmd": AjaxResponse.APPEND, "data": "some value", "id": "#team"}],
        ),
        (
            "assign",
            {"element": "#team", "value": "some value"},
            [{"cmd": AjaxResponse.ASSIGN, "data": "some value", "id": "#team"}],
        ),
        (
            "redirect",
            {"url": "https://some.address.co.uk"},
            [
                {
                    "cmd": AjaxResponse.REDIRECT,
                    "data": None,
                    "url": "https://some.address.co.uk",
                }
            ],
        ),
        (
            "reload_location",
            None,
            [{"cmd": AjaxResponse.RELOAD_LOCATION, "data": None}],
        ),
        (
            "remove",
            {"element": "#element_id"},
            [{"cmd": AjaxResponse.REMOVE, "data": None, "id": "#element_id"}],
        ),
        (
            "add_class",
            {"element": "#element_id", "class_name": "some-class"},
            [
                {
                    "cmd": AjaxResponse.ADD_CLASS,
                    "data": "some-class",
                    "id": "#element_id",
                }
            ],
        ),
        (
            "remove_class",
            {"element": "#element_id", "class_name": "some-class"},
            [
                {
                    "cmd": AjaxResponse.REMOVE_CLASS,
                    "data": "some-class",
                    "id": "#element_id",
                }
            ],
        ),
        (
            "set_class",
            {"element": "#element_id", "class_name": "some-class"},
            [
                {"cmd": AjaxResponse.REMOVE_CLASS, "data": None, "id": "#element_id"},
                {
                    "cmd": AjaxResponse.ADD_CLASS,
                    "data": "some-class",
                    "id": "#element_id",
                },
            ],
        ),
        (
            "return_json",
            {"data": {"class_name": "some-class"}},
            {"class_name": "some-class"},
        ),
        (
            "return_json",
            {"data": {"errors": {"msg": "some-class"}}},
            {"msg": "some-class"},
        ),
        (
            "script",
            {"javascript": "some javascript"},
            [{"cmd": AjaxResponse.RUN_SCRIPT, "data": "some javascript"}],
        ),
        (
            "show",
            {"element": "#element"},
            [{"cmd": AjaxResponse.SHOW, "data": None, "id": "#element"}],
        ),
        (
            "hide",
            {"element": "#element"},
            [{"cmd": AjaxResponse.HIDE, "data": None, "id": "#element"}],
        ),
        (
            "insert_before",
            {"element": "#element", "value": "Some value"},
            [
                {
                    "cmd": AjaxResponse.INSERT_BEFORE,
                    "data": "Some value",
                    "id": "#element",
                }
            ],
        ),
        (
            "init_ajax_links",
            None,
            [{"cmd": AjaxResponse.INIT_AJAX_LINKS, "data": None}],
        ),
        (
            "init_ajax_select",
            None,
            [{"cmd": AjaxResponse.INIT_AJAX_SELECT, "data": None}],
        ),
        (
            "init_ajax_forms",
            None,
            [{"cmd": AjaxResponse.INIT_AJAX_FORMS, "data": None}],
        ),
        (
            "set_input_value",
            {"element": "#element_id", "value": "some value"},
            [
                {
                    "cmd": AjaxResponse.SET_INPUT_VALUE,
                    "data": "some value",
                    "id": "#element_id",
                }
            ],
        ),
        (
            "modal",
            {"element": "#element_id", "action": "some value"},
            [{"cmd": AjaxResponse.MODAL, "data": "some value", "id": "#element_id"}],
        ),
        (
            "set_url",
            {"element": "#element_id", "url": "https://some-url.com"},
            [
                {
                    "cmd": AjaxResponse.URL,
                    "data": "https://some-url.com",
                    "id": "#element_id",
                }
            ],
        ),
        (
            "set_form_action",
            {"element": "#element_id", "action": "https://some-url.com"},
            [
                {
                    "cmd": AjaxResponse.SET_FORM_ACTION,
                    "data": "https://some-url.com",
                    "id": "#element_id",
                }
            ],
        ),
        (
            "load_script",
            {"name": "script name", "callback": "callback function"},
            [
                {
                    "cmd": AjaxResponse.LOAD_SCRIPT,
                    "data": None,
                    "script": "script name",
                    "callback": "callback function",
                }
            ],
        ),
    )

    @data_provider(datasource_for_tests)
    def test_all(self, command: str, parameters: dict or None, expected: list):
        function_name = getattr(AjaxResponse(), command)
        response: AjaxResponse = (
            function_name(**parameters) if parameters else function_name()
        )
        self.assertEqual(response.get_list(), expected)
        self.assertEqual(len(response.get_json()), len(json.dumps(expected)))
        self.assertEqual(len(str(response)), len(json.dumps(expected)))

    def test_print_output(self):
        given = {"name": "script name", "callback": "callback function"}
        expected = (
            json.dumps(
                [
                    {
                        "cmd": AjaxResponse.LOAD_SCRIPT,
                        "data": None,
                        "script": "script name",
                        "callback": "callback function",
                    }
                ]
            )
            + "\n"
        )

        response: AjaxResponse = AjaxResponse().load_script(**given)

        capturedOutput = StringIO()
        sys.stdout = capturedOutput
        response.print_output()
        sys.stdout = sys.__stdout__

        self.assertEqual(len(capturedOutput.getvalue()), len(expected))



"""copyright (c) 2014 - 2024 Beeflow Ltd.

Author Rafal Przetakowski <rafal.p@beeflow.co.uk>"""

from .ajax_request import AjaxFormRequest
from .ajax_response import AjaxResponse
from .websocket_response import WebSocketResponse

__all__ = [
    "AjaxResponse",
    "AjaxFormRequest",
    "WebSocketResponse",
]



"""copyright (c) 2014 - 2024 Beeflow Ltd.

Author Rafal Przetakowski <rafal.p@beeflow.co.uk>"""
import json
from typing import Any, Optional, Self

from beeflow_ajax.lib import AjaxResponse


class WebSocketResponse(AjaxResponse):
    INIT_WEBSOCKET_FORMS = "initWebsocketForms"

    async def send_json(
        self, v: Optional[dict[str, Any] | list[dict[str, Any]]] = None
    ):
        """This command is available only for the WebSockets."""
        if isinstance(v, dict):
            v = [v]

        self.commands.append(v or [])

        response = json.dumps((self.commands or []))
        self.commands = []

        await self.response_.send_json(response)

    def init_websocket_forms(self) -> Self:
        """Initialize WebSockets forms."""
        self._add_command(self.INIT_WEBSOCKET_FORMS)
        return self



"""copyright (c) 2014 - 2024 Beeflow Ltd.

Author Rafal Przetakowski <rafal.p@beeflow.co.uk>"""

import json
from typing import Any, Optional

from pydantic.v1 import BaseModel, Extra


def to_camel_case(snake_str):
    return "".join(x.capitalize() for x in snake_str.lower().split("_"))


def to_lower_camel_case(snake_str):
    # We capitalize the first letter of each component except the first one
    # with the 'capitalize' method and join them together.
    camel_string = to_camel_case(snake_str)
    return snake_str[0].lower() + camel_string[1:]


base_model_attributes_map: dict[str, str] = {"class_name": "class"}


class BaseHTMLModel(BaseModel, extra=Extra.allow):
    element_type: str = "element"
    element_name: str

    class_name: Optional[str] = None
    inner_text: Optional[str] = None
    inner_element: Optional["BaseHTMLModel"] = None
    inner_elements: Optional[list["BaseHTMLModel"]] = None

    append_text: Optional[str] = None

    def model_dump_json(self, **kwargs) -> str:
        return json.dumps(self.model_dump())

    def model_dump(self, **kwargs) -> dict[str, Any]:
        __all: dict = super().dict(by_alias=True, **kwargs)
        result: dict = self.dump_without_unset(__all.items())

        return result

    def dump_without_unset(self, items: dict) -> dict:
        result: dict = {}

        for key, value in items:
            if not value:
                continue

            if key in base_model_attributes_map.keys():
                key = base_model_attributes_map[key]

            if key.startswith("data_"):
                key = key.replace("_", "-")

            key = to_lower_camel_case(key)

            if isinstance(value, list):
                inner_elements: list = []

                for inner_value in value:
                    inner_elements.append(self.dump_without_unset(inner_value.items()))

                result.update({key: inner_elements})

                continue

            if isinstance(value, BaseModel):
                result.update({key: value.model_dump()})

            if isinstance(value, dict):
                result.update({key: self.dump_without_unset(value.items())})
                continue

            result.update({key: value})

        return result



"""copyright (c) 2014 - 2024 Beeflow Ltd.

Author Rafal Przetakowski <rafal.p@beeflow.co.uk>"""
import ast
from urllib.parse import unquote


class AjaxFormRequest:
    def __init__(self, request):
        data_str = unquote(request.body.decode()).replace("data=", "").replace("+", " ")
        self._request_data: dict = ast.literal_eval(data_str)

    @property
    def data(self) -> dict:
        return self._request_data



"""copyright (c) 2014 - 2024 Beeflow Ltd.

Author Rafal Przetakowski <rafal.p@beeflow.co.uk>"""

from pydantic import Field

from beeflow_ajax.lib.html_models.base_html_model import BaseHTMLModel


class PaginationButton(BaseHTMLModel):
    element_name: str = "li"
    class_name: str = "page-item"

    disabled: bool = Field(exclude=True)

    def __init__(self, **data):
        super().__init__(**data)

        if self.disabled:
            self.class_name += " disabled"



"""Beeflow Ajax helps communicate HTML service with backend.

@author Rafal Przetakowski <rafal.p@beeflow.co.uk>"""

import json
from typing import Any, Callable, Dict, List, Optional, Self, TypeVar

from beeflow_ajax.lib.html_models.base_html_model import BaseHTMLModel

T = TypeVar('T', bound='AjaxResponse')


class AjaxCommandHandlerError(Exception):
    ...


class AjaxResponse:
    """Response object for AJAX."""

    ALERT = "alert"
    ALERT_SUCCESS = "alertSuccess"
    ALERT_ERROR = "alertError"
    ALERT_WARNING = "alertWarning"
    ALERT_INFO = "alertInfo"
    DEBUG = "debug"
    APPEND = "append"
    ASSIGN = "assign"
    APPEND_ELEMENT = "appendElement"
    APPEND_ELEMENTS = "appendElements"
    ASSIGN_ELEMENT = "assignElement"
    ASSIGN_ELEMENTS = "assignElements"
    APPEND_LIST = "appendList"
    ASSIGN_LIST = "assignList"
    REDIRECT = "redirect"
    RELOAD_LOCATION = "reloadLocation"
    REMOVE = "remove"
    ADD_CLASS = "addClass"
    REMOVE_CLASS = "removeClass"
    RUN_SCRIPT = "runScript"
    SHOW = "show"
    HIDE = "hide"
    INSERT_BEFORE = "insertBefore"
    INSERT_AFTER = "insertAfter"
    INIT_AJAX_LINKS = "initAjaxLinks"
    INIT_AJAX_SELECT = "initAjaxSelect"
    INIT_AJAX_FORMS = "initAjaxForms"
    LOAD_SCRIPT = "loadScript"
    SET_INPUT_VALUE = "setInputValue"
    MODAL = "modal"
    URL = "setUrl"
    SET_FORM_ACTION = "setFormAction"
    SET_ATTRIBUTE = "setAttribute"
    ROW_UP = "rowUp"
    ROW_DOWN = "rowDown"
    FORM_FIELD_ERROR = "formFieldError"

    _instance = None
    _is_initialized = False

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(AjaxResponse, cls).__new__(cls)
        return cls._instance

    def __init__(self, response=None):
        """Initializes a singleton instance, only once."""
        if not self._is_initialized:
            self.commands = []
            self.response_handler = response
            self.command_handlers = {}
            self._is_initialized = True

    def __str__(self):
        """Returns json string."""
        return self.get_json()

    def response(self, *args, **kwargs):
        """
        Method returns response as the:
            * response (ex. HTTPResponse)
            * dictionary if we didn't pass any response object to the AjaxResponse initializer
        """
        commands = self.commands or {}
        self.commands = []

        if not self.response_handler:
            return commands

        return self.response_handler(commands, *args, **kwargs)

    def get_json(self) -> str:
        """Returns json string."""
        commands = self.commands or {}
        return json.dumps(commands)

    def print_output(self):
        """Prints json string."""
        print(self.get_json())

    def get_list(self) -> List:
        """Returns commands as a list."""
        return self.commands

    def alert(self, msg: str) -> T:
        """Prepares alert command."""
        self._add_command(self.ALERT, {}, str(msg))
        return self

    def alert_success(self, msg: str, title: str = "", callback: Optional[str] = None) -> T:
        """Prepares success message command."""
        self._add_command(
            self.ALERT_SUCCESS, {"title": title, "callback": callback}, str(msg)
        )
        return self

    def alert_error(self, msg: str, title: str = "", callback: Optional[str] = None) -> T:
        """Prepares error message command."""
        self._add_command(
            self.ALERT_ERROR, {"title": title, "callback": callback}, str(msg)
        )
        return self

    def alert_warning(self, msg: str, title: str = "", callback: Optional[str] = None) -> T:
        """Prepares warning message command."""
        self._add_command(
            self.ALERT_WARNING, {"title": title, "callback": callback}, str(msg)
        )
        return self

    def alert_info(self, msg: str, title: str = "", callback: Optional[str] = None) -> T:
        """Prepares info message command."""
        self._add_command(
            self.ALERT_INFO, {"title": title, "callback": callback}, str(msg)
        )
        return self

    def debug(self, data) -> T:
        """Prepares debug command."""
        self._add_command(self.DEBUG, {}, data)
        return self

    def append_element(self, element: str, element_data: dict[str:Any] | BaseHTMLModel) -> T:
        if isinstance(element_data, BaseHTMLModel):
            element_data = element_data.model_dump()

        self._add_command(self.APPEND_ELEMENT, {"id": element, "element": element_data})
        return self

    def append_elements(self, element: str, element_data: list[dict[str:Any] | BaseHTMLModel]) -> T:
        if isinstance(element_data[0], BaseHTMLModel):
            element_data = [edata.model_dump() for edata in element_data]

        self._add_command(self.APPEND_ELEMENTS, {"id": element, "elements": element_data})
        return self

    def assign_element(self, element: str, element_data: dict[str:Any] | BaseHTMLModel) -> T:
        if isinstance(element_data, BaseHTMLModel):
            element_data = element_data.model_dump()

        self._add_command(self.ASSIGN_ELEMENT, {"id": element, "element": element_data})
        return self

    def assign_elements(self, element: str, element_data: list[dict[str:Any] | BaseHTMLModel]) -> T:
        if isinstance(element_data[0], BaseHTMLModel):
            element_data = [edata.model_dump() for edata in element_data]

        self._add_command(self.ASSIGN_ELEMENTS, {"id": element, "elements": element_data})
        return self

    def append_list(self, element: str, element_data: list[dict[str:Any]], list_type: str) -> T:
        self._add_command(
            self.APPEND_LIST,
            {"id": element, "element": element_data, "list_type": list_type},
        )
        return self

    def assign_list(self, element: str, element_data: list[dict[str:Any]], list_type: str) -> T:
        self._add_command(
            self.ASSIGN_LIST,
            {"id": element, "element": element_data, "list_type": list_type},
        )
        return self

    def append(self, element: str, value: str) -> T:
        """Prepares append command which adds value to element.

        The element can be #id, .class or just tag ex. p
        """
        self._add_command(self.APPEND, {"id": element}, value)
        return self

    def assign(self, element: str, value: str) -> T:
        """Prepares assign command which adds value to element.

        The element can be #id, .class or HTML tag.
        """
        self._add_command(self.ASSIGN, {"id": element}, value)
        return self

    def redirect(self, url: str) -> T:
        """Redirect command for url redirection."""
        self._add_command(self.REDIRECT, {"url": url})
        return self

    def reload_location(self) -> T:
        """Reload location command."""
        self._add_command(self.RELOAD_LOCATION)
        return self

    def remove(self, element: str) -> T:
        """Removes element by #id, .class or HTML tag."""
        self._add_command(self.REMOVE, {"id": element})
        return self

    def add_class(self, element: str, class_name: str) -> T:
        """Adds class to element which can be #id, .class or HTML tag."""
        self._add_command(self.ADD_CLASS, {"id": element}, class_name)
        return self

    def remove_class(self, element: str, class_name: str = None) -> T:
        """Removes class from element which can be #id, .class or HTML tag."""
        self._add_command(self.REMOVE_CLASS, {"id": element}, class_name)
        return self

    def set_class(self, element: str, class_name: str) -> T:
        """Sets new class on element which can be #id, .class or HTML tag."""
        self.remove_class(element)
        self.add_class(element, class_name)
        return self

    def return_json(self, data: Dict) -> T:
        """Allows return json as a command."""
        try:
            self.commands = data["errors"]
        except KeyError:
            self.commands = data
        return self

    def script(self, javascript: str) -> T:
        """Allows send javascript script to frontend."""
        self._add_command(self.RUN_SCRIPT, {}, javascript)
        return self

    def show(self, element: str) -> T:
        """Shows element."""
        self._add_command(self.SHOW, {"id": element})
        return self

    def hide(self, element: str) -> T:
        """Hides element."""
        self._add_command(self.HIDE, {"id": element})
        return self

    def insert_before(self, element: str, value: str) -> T:
        """Inserts value before element."""
        self._add_command(self.INSERT_BEFORE, {"id": element}, value)
        return self

    def insert_after(self, element: str, value: str) -> T:
        """Inserts value after element."""
        self._add_command(self.INSERT_AFTER, {"id": element}, value)
        return self

    def init_ajax_links(self) -> T:
        """Initialize ajax links."""
        self._add_command(self.INIT_AJAX_LINKS)
        return self

    def init_ajax_select(
            self, callback: Optional[str] = None, callbackParams: Optional[dict] = None,
            callbackCommands: Optional[list] = None
    ) -> T:
        """Initialize ajax select."""

        self._add_command(
            self.INIT_AJAX_SELECT,
            {"callback": callback, "callbackParams": callbackParams, "callbackCommands": callbackCommands}
        )
        return self

    def init_ajax_forms(self) -> T:
        """Initialize ajax forms."""
        self._add_command(self.INIT_AJAX_FORMS)
        return self

    def load_script(self, name: str, callback: str) -> T:
        """Allows load javascript script from file."""
        self._add_command(self.LOAD_SCRIPT, {"script": name, "callback": callback})
        return self

    def set_input_value(self, element: str, value: str) -> T:
        self._add_command(self.SET_INPUT_VALUE, {"id": element}, value)
        return self

    def modal(self, element: str, action: str) -> T:
        self._add_command(self.MODAL, {"id": element}, action)
        return self

    def set_url(self, element: str, url: str) -> T:
        self._add_command(self.URL, {"id": element}, url)
        return self

    def set_form_action(self, element: str, action: str) -> T:
        self._add_command(self.SET_FORM_ACTION, {"id": element}, action)
        return self

    def set_attribute(self, element: str, attribute_name: str, value: Any) -> T:
        self._add_command(
            self.SET_ATTRIBUTE,
            {"id": element, "attribute": attribute_name, "value": value},
        )
        return self

    def row_up(self, element: str) -> T:
        """Move table row (or other element) up."""
        self._add_command(self.ROW_UP, {"id": element})
        return self

    def row_down(self, element: str) -> T:
        """Move table row (or other element) down."""
        self._add_command(self.ROW_DOWN, {"id": element})
        return self

    def form_field_error(self, element: str, error_message: str) -> T:
        self._add_command(self.FORM_FIELD_ERROR, {"id": element, "error_message": error_message})
        return self

    def add_command(self, command: str, handler: Callable[[Dict[str, Any]], T]) -> None:
        """Registers a new command handler dynamically."""
        self.command_handlers[command] = handler

    def call_command(self, command: str, details: Dict[str, Any]) -> T:
        """Adds a command to the command list and attempts to process it dynamically."""
        handler = self.command_handlers.get(command, None)
        if handler:
            return handler(details)

        raise AjaxCommandHandlerError(f"No handler registered for command: {command}")

    def _add_command(self, command: str, attributes: Dict = None, m_data=None) -> T:
        """Adds command."""
        if attributes is None:
            attributes = {}

        attributes["cmd"] = command
        attributes["data"] = m_data
        self.commands.append(attributes)
        return self



"""copyright (c) 2014 - 2024 Beeflow Ltd.

Author Rafal Przetakowski <rafal.p@beeflow.co.uk>"""

from beeflow_ajax.lib.html_models.base_html_model import BaseHTMLModel


class Icon(BaseHTMLModel):
    element_name: str = "i"



"""copyright (c) 2014 - 2024 Beeflow Ltd.

Author Rafal Przetakowski <rafal.p@beeflow.co.uk>"""

import os

import setuptools

os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setuptools.setup(
    name="beeflow-ajax",
    version="1.0.17",
    author="Rafal Przetakowski",
    author_email="office@beeflow.co.uk",
    description="Library to maintain ajax and websockets communication without writing complicated code in JS",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    url="https://github.com/beeflow/BeeflowAjaxPy",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.11",
)



"""copyright (c) 2014 - 2024 Beeflow Ltd.

Author Rafal Przetakowski <rafal.p@beeflow.co.uk>"""
from beeflow_ajax.lib.html_models.base_html_model import BaseHTMLModel


class AjaxPaginatorUrl(BaseHTMLModel):
    element_name: str = "a"
    class_name: str = "page-link ajax-link"
    href: str
